<?php

namespace onespace\tools\rabbit\migrations;

use yii\db\Migration;

class m250514_080110_increate_ttl_rabbit_outbox extends Migration {
    /**
     * {@inheritdoc}
     */
    public function safeUp() {
        $this->alterColumn(
            '{{%rabbit_outbox}}',
            'ttl',
            $this->bigInteger()->unsigned()
        );
    }

    /**
     * {@inheritdoc}
     */
    public function safeDown() {
        $this->alterColumn(
            '{{%rabbit_outbox}}',
            'ttl',
            $this->integer()
        );
    }

    /*
    // Use up()/down() to run migration code without a transaction.
    public function up()
    {

    }

    public function down()
    {
        echo "m250514_080110_increate_ttl_rabbit_outbox cannot be reverted.\n";

        return false;
    }
    */
}
