<?php

namespace onespace\tools\rabbit\models;

use yii\helpers\ArrayHelper;

/**
 * This is the model class for table "rabbit_outbox".
 */
class RabbitOutbox extends \onespace\tools\rabbit\models\base\RabbitOutbox {
    public function behaviors() {
        return ArrayHelper::merge(
            parent::behaviors(),
            [
                # custom behaviors
            ]
        );
    }

    public function rules() {
        return ArrayHelper::merge(
            parent::rules(),
            [
                # custom validation rules
            ]
        );
    }
}
