<?php

namespace onespace\tools\rabbit;

use yii\base\BootstrapInterface;

/**
 * Bootstrap
 * @package onespace/tools/rabbit
 *
 * @see https://www.yiiframework.com/doc/guide/2.0/en/structure-extensions#bootstrapping-classes
 */
class RabbitBootstrap implements BootstrapInterface {
    /**
     * Bootstrap method to be called during application bootstrap stage.
     *
     * @param Application $app the application currently running
     */
    public function bootstrap($app) {
        \Yii::setAlias('@onespace/tools/rabbit', __DIR__);
    }
}
