<?php

return [
    'class' => \onespace\tools\rabbit\components\Rabbit::class,
    'serverUrl' => getenv('RABBIT_HOST_URL'),
    'serverPort' => getenv('RABBIT_HOST_PORT') ? getenv('RABBIT_HOST_PORT') : 5672,
    'vHost' => getenv('RABBIT_VHOST') ? getenv('RABBIT_VHOST') : 'dev',
    'username' => getenv('RABBIT_USERNAME'),
    'password' => getenv('RABBIT_PASSWORD'),
    'outboxModel' => \onespace\tools\rabbit\models\RabbitOutbox::class,
];
