<?php

namespace onespace\tools\rabbit\services\interfaces;

use Interop\Amqp\Impl\AmqpMessage;

interface ServiceInterface {
    /**
     * Class constructor. The received Rabbit message is decoded here.
     */
    public function __construct(AmqpMessage $message);

    /**
     * Executes the various tasks based on the send message.
     *
     * @return  bool
     *
     * @access  public
     */
    public function run(): bool;
}
