<?php
// This class was automatically generated by a giiant build task
// You should not change it manually as it will be overwritten on next build

namespace onespace\tools\rabbit\models\base;

use Yii;
use yii\helpers\ArrayHelper;

/**
 * This is the base-model class for table "rabbit_outbox".
 *
 * @property integer $id
 * @property string $worker_class
 * @property string $message_body
 * @property string $message_headers
 * @property string $routing_key
 * @property string $date_created
 */
abstract class RabbitTransactionErrorLog extends \yii\db\ActiveRecord {

    /**
     * @inheritdoc
     */
    public static function tableName() {
        return 'rabbit_transaction_error_log';
    }

    /**
     * @inheritdoc
     */
    public function rules() {
        $parentRules = parent::rules();
        return ArrayHelper::merge($parentRules, [
            [['worker_class'], 'string', 'max' => 200],
            [['message_body', 'message_headers', 'routing_key'], 'string'],
            [['date_created'], 'required'],
            [['date_created'], 'safe'],
        ]);
    }

    /**
     * @inheritdoc
     */
    public function attributeLabels() {
        return ArrayHelper::merge(parent::attributeLabels(), [
            'id' => 'ID',
            'worker_class' => 'Worker Class',
            'message_body' => 'Message Body',
            'message_headers' => 'Message Headers',
            'routing_key' => 'Routing Key',
            'date_created' => 'Date ',
        ]);
    }
}
